/**
 * orbit.java - an orbit simulator applet
 * By Bob Jenkins, July & August of 1998
 *
 * Recognized applet parameters:
 * "background": the background color, eg "000000"
 * "increment": how big a step to take between scenes
 * "sun": objects only interact with the sun, not with each other.  
 *    "moon1" is the sun.  This reduces the calculations from O(n*n)
 *    to O(n).  Any value turns this on, eg: "yes"
 * "stop": any parameter causes applet to start off not moving
 * "length": length of simulation, stop after this point.
 * "trail": leave trails; don't erase old positions of moons
 * "moons": an integer, the number of moons, eg "3"
 * "moon1", "moon2", however many "moons" specified:
 *    position, velocity, mass, color, size.  For example:
 *   "p(0.0, 0.0, 0.0) v(0.0, 0.0, 0.0) 100.0 ffffff 10.0"
 * "eye": how the camera is set up to view this,
 *    position, left(radians), up(radians), clockwise(radians), zoom
 *    "100.0 0.0 0.0 0.0 100.0"
 */
import java.applet.*;
import java.awt.*;
import java.lang.Math;

// An applet that controls an eye which views a set of moons
public final class orbit extends Applet implements Runnable {
  double increment;       // how much to increment time each step
  double         t;       // time
  double         maxt;    // end of time;
  private Thread runner;  // background thread timing motions
  eye            camera;  // camera for viewing the moons
  simulate       sim;     // all the moons in the system
  Color          bgcolor; // background color
  boolean   should_move;  // should the objects be moving
  Image     buffered;     // double buffering buffer
  Graphics  bg;           // buffered.getGraphics()
  double    oldmousex;    // last handled mouse x
  double    oldmousey;    // last handled mouse y
  double    mousex;       // last recorded mouse x
  double    mousey;       // last recorded mouse y
  boolean   drag;         // is the mouse draging the image?
  boolean   xymode;       // true is left/up, false is zoom/clockwise
  boolean   trail;        // leave trails?
  
  public void init() {
    // set up infrastructure
    String param;
    this.t = 0.0;

    // get parameters
    this.bgcolor = point.s2c(getParameter("background"), Color.black);

    this.increment = point.s2d(getParameter("increment"), 0.1);

    this.maxt = point.s2d(getParameter("length"), 1.0e30)/this.increment;

    this.should_move = (getParameter("stop") == null);

    this.trail = (getParameter("trail") != null);

    // all the objects in the system
    param = getParameter("moons");
    moon luna[];
    if (param==null) {
      luna = new moon[1];
      luna[0] = new moon("p(0.0,0.0,0.0) v(0.0,0.0,0.0) 1000.0 00ff00 15.0");
    } else {
      int i=0;
      luna = new moon[Integer.parseInt(param)];
      for (int j=0; i<luna.length; ++j) {
	param = getParameter("moon"+j);
	if (j > 2*luna.length) break;
	if (param==null) continue;
	luna[i] = new moon(param);
	++i;
      }
      moon luna2[] = new moon[i];
      for (int j=0; j<i; ++j) luna2[j] = luna[j];
      luna = luna2;
    }

    // In "sun" mode, only the first object attracts.  (Runs faster.)
    param = getParameter("sun");
    if (param!=null) {
      luna[0].other = luna;
      for (int i=1; i<luna.length; ++i) {
	luna[i].other = new moon[1];
	luna[i].other[0] = luna[0];
      }
    } else {
      for (int i=0; i<luna.length; ++i) {
	luna[i].other = luna;
      }
    }

    // the eye watching this world
    param = getParameter("eye");
    if (param==null) 
      param = "10.0  0.0  0.0  0.0  20.0";
    this.camera = new eye(param, size().width/2, size().height/2);
    this.drag = false;
    this.oldmousex = 0.0;
    this.oldmousey = 0.0;
    this.mousex = 0.0;
    this.mousey = 0.0;
    this.sim = new simulate(luna, this.increment);
  }

  void startstop() {
    if (should_move && runner==null) {
      if (t > maxt) { init(); }
      runner = new Thread(this);
      runner.start();
    } else if (!should_move && runner != null && runner.isAlive()) {
      runner.stop();
      runner = null;
    }
  }

  void clear() {
    bg.setColor(bgcolor);
    bg.fillRect(0, 0, size().width, size().height);
    bg.setColor(bgcolor.gray);
    bg.drawLine(0, 0, size().width, 0);
    bg.drawLine(0, 0, 0, size().height);
    bg.drawLine(size().width-1, 0, size().width-1, size().height-1);
    bg.drawLine(0, size().height-1, size().width-1, size().height-1);
  }

  public void start() {
    // if objects should be moving, start them moving again
    startstop();

    if (buffered == null) {
      buffered = createImage(size().width, size().height);
      bg = buffered.getGraphics();
      clear();
    }

    // register that the screen needs painting
    repaint();
  }

  public void stop() {
    runner = null;

    if (bg != null) {
      bg.dispose();
      bg = null;
    }
    if (buffered != null) {
      buffered = null;
    }
  }

  // mouse clicks turn the action on/off
  public boolean mouseUp(Event e, int x, int y) {
    if (!drag) {
      should_move = (!should_move);
      startstop();
    }
    drag = true;
    return true;
  }

  // mouse clicks turn the action on/off
  public boolean mouseDown(Event e, int x, int y) {
    mousex = (double)(x-this.camera.centerx);
    mousey = (double)(y-this.camera.centery);
    oldmousex = mousex;
    oldmousey = mousey;
    xymode = (4.0*(mousex*mousex + mousey*mousey) <
	      (double)Math.min(this.camera.centerx*this.camera.centerx, 
			       this.camera.centery*this.camera.centery));
    drag = false;
    return true;
  }

  // mouse clicks turn the action on/off
  public boolean mouseDrag(Event e, int x, int y) {
    mousex = (double)(x-this.camera.centerx);
    mousey = (double)(y-this.camera.centery);
    return true;
  }

  public void run() {
    if (bg != null) {
      clear();
    }
    while (runner != null) {
      repaint();    // notify paint() that it needs to run
      try {Thread.sleep(50);} catch (InterruptedException e) {}
      if (t > maxt) {
	should_move = false;
	runner = null;
      }
    }
  }

  // called by system in a separate thread when repaint() has been called.
  // Beware, several repaint()s may cause only one update() to happen.
  public void update(Graphics g) {
    paint(g);
  }

  public void paint(Graphics g) {
    if (bg == null) return;     // there seem to be some race conditions

    // erase old image
    if (!trail) {
      bg.setColor(bgcolor);
      sim.clear(bg);
    }

    // find time increment
    this.t += 1.0;

    // compute new positions
    sim.move();
    
    // update camera
    double myx = mousex;
    double myy = mousey;
    double deltamousex = myx - oldmousex;
    double deltamousey = myy - oldmousey;
    if (deltamousex != 0 || deltamousey != 0) {
      drag = true;
      if (xymode) {
	camera.left(deltamousex*Math.abs(deltamousex)/camera.magnification);
	camera.up(deltamousey*Math.abs(deltamousey)/camera.magnification);
      } else {
	double mynorm = oldmousex*oldmousex + oldmousey*oldmousey;
	if (mynorm > 0.5) {
	  double myzoom = Math.sqrt((myx*myx + myy*myy) / mynorm);
	  camera.zoom(myzoom);
	  myx /= myzoom;
	  myy /= myzoom;
	  double myrot = Math.asin((oldmousey*myx - oldmousex*myy) / mynorm);
	  camera.clockwise(myrot);
	}
      }
      oldmousex += (int)deltamousex;
      oldmousey += (int)deltamousey;
      if (trail) {
	clear();
      }
    }

    // draw the moons in their new positions
    sim.draw(bg, camera);

    // actually display the new image
    this.getGraphics().drawImage(buffered, 0, 0, null);
  }
}
