/**
 * simulate.java -- Simulate a set of moons
 *
 * By Bob Jenkins, July 1998
 * Permission granted to use, reuse, rewrite this without notifying me.
 */

import java.awt.*;
import java.lang.Math;
import java.util.*;

public final class simulate {
  moon   luna[];  // the set of moons being simulated
  point  temp;    // working memory;

  public simulate(moon luna[], double increment) {
    this.temp = new point(0.0, 0.0, 0.0);
    this.luna = luna;

    // scale the masses so that delta time is 1.0
    for (int i=0; i<luna.length; ++i) {
      luna[i].m *= (increment*increment*0.04);
      luna[i].v.scale(increment*0.2);
    }

    // put center of gravity, net velocity at (0,0,0)
    center();

    // we really wanted velocity = p1 - p0, not actual velocity at p1.
    for (int i=0; i<luna.length; ++i) {
      point.plusax(luna[i].v, luna[i].v, -0.5, luna[i].a);
    }
  }

  // make the center of gravity and net momentum (0,0,0)
  final private void center() {
    double m = 0.0;  // total mass
    point  p = new point(0.0, 0.0, 0.0);  // center of gravity so far
    point  v = new point(0.0, 0.0, 0.0);  // mv is total momentum so far
    for (int i=0; i<luna.length; ++i) {
      moon q = luna[i];
      m += q.m;
      point.plusax(v, v, q.m, q.v);
      point.plusax(p, p, q.m, q.p);
    }
    for (int i=0; i<luna.length; ++i) {
      moon q = luna[i];
      point.plusax(q.v, q.v, -1.0/m, v);
      point.plusax(q.p, q.p, -1.0/m, p);
    }
  }

  // Would you believe this is a cubic approximation of the right formula?
  // The formula this is based on is
  //   p1 = p0 + Dp0 + a0 + D^2a0/12 + D^3a0/12 + ...
  // where the constants are 1, 1, 1, 0, 1/12, 1/12, 19/240, 3/40, 863/12096
  // 275/4032, 33953/518400, and so on near .065 and slowly decreasing.  
  // D^ix1 is D^(i-1)x1 - D^(i-1)x0, and D^0x1 = x1.
  // This is loosely based on the adams-moulton method, but I had to figure 
  // out the constants myself.  n terms gives exact results when approximating
  // an n-1 degree polynomial.
  //
  // The first three terms produces orbits even more stable than the full
  // chain.  So why not use just the first three terms?
  //
  final private void adams_jenkins() {
    // Find acceleration.  Find it for all points before moving any of them.
    accel();
    // Move the points to p_1 = p_0 + (p0-p_-1) + a_0.
    for (int i=0; i<luna.length; ++i) {
      moon  q = luna[i];
      point.plusx(q.v, q.v, q.a);  // new (p1-p0) (NOT velocity at p0)
      point.plusx(q.p, q.p, q.v);  // new p
    }
  }

  // move all moons forward by one time increment
  final public void move() {
    adams_jenkins();
    adams_jenkins();
    adams_jenkins();
    adams_jenkins();
    adams_jenkins();
  }

  // determine the acceleration
  private void accel() {
    for (int i=0; i<luna.length; ++i) {
      moon q = luna[i];
      q.a.zero();
      temp.zero();

      // include the effects of acceleration from all the others
      for (int j=0; j<q.other.length; ++j) {
	moon q2 = q.other[j];
	if (q == q2)
	  continue;
	temp.minusx(temp, q2.p, q.p);
	double scale = temp.dot(temp);
	double dist = Math.sqrt(scale);
	if (scale <= 0.0) {
	} else {
	  scale = q2.m/(scale*dist);
	  point.plusax(q.a, q.a, scale, temp);
	}
      }
    }
  }

  // erase all the current moons
  public void clear(Graphics bg) {
    for (int i=0; i<luna.length; ++i) {
      int q = luna[i].screenr;
      if (q > 2) {
	bg.fillOval(luna[i].screenx, luna[i].screeny, q, q);
      } else {
	bg.fillRect(luna[i].screenx, luna[i].screeny, 
		    luna[i].screenx+q, luna[i].screeny+q);
      }
    }
  }

  // draw new moons
  public void draw(Graphics bg, eye camera) {

    // Find new positions in screen coordinates
    for (int i=0; i<luna.length; ++i) {
      point temp = luna[i].peye;
      camera.map(temp, luna[i].p);
      if (temp.z > 0.0) {
	int q = (int)(camera.magnification*luna[i].r/temp.z);
	if (q < 1) q = 1;
	luna[i].screenr = q;
	luna[i].screenx = camera.mapx(temp)-q/2;
	luna[i].screeny = camera.mapy(temp)-q/2;
      } else {
	luna[i].screenx = luna[i].screeny=luna[i].screenr = 0;
      }
    }

    // kinda sort so closest moons are drawn last
    // assumes moons don't move very fast
    for (int i=luna.length; --i > 0;)
      if (luna[i].peye.z > luna[i-1].peye.z) {
	moon temp = luna[i];   luna[i] = luna[i-1];   luna[i-1] = temp;
      }

    // draw the moons
    for (int i=0; i<luna.length; ++i) {
      int q = luna[i].screenr;
      if (q > 2) {
	bg.setColor(luna[i].c);
	bg.fillOval(luna[i].screenx, luna[i].screeny, q, q);
      } else if (q > 0) {
	bg.setColor(luna[i].c);
	bg.fillRect(luna[i].screenx, luna[i].screeny, q, q);
      }
    }
  }
}
